"""
Ctypes constants module for the SIE / IFM CANfox interface

Copyright (C) 2022 Matt Woodhead
"""

from ctypes import (
    c_double,
    c_int,
    c_long,
    c_ulong,
    c_longlong,
    c_ulonglong,
    c_wchar_p,
    c_bool,
    c_ubyte,
    POINTER,
)
import sys

from can.ctypesutil import HANDLE
from .structures import (
    CANMsgStruct,
    CANMsgBuffer,
    CANStatusStruct,
    CANIDStatusStruct,
    CANCounterStruct,
    CANCounterStruct2,
    CANLevelHistStruct,
    CANBusLoadStruct,
    CANInstalledDevicesStruct,
)


IS_PYTHON_64BIT = sys.maxsize > 2**32


NTCAN_SUCCESS = 0
NTCAN_RX_TIMEOUT = -1
NTCAN_TX_TIMEOUT = -2
NTCAN_CONTR_BUSOFF = -3
NTCAN_NO_ID_ENABLED = -4
NTCAN_ID_ALREADY_ENABLED = -5
NTCAN_ID_NOT_ENABLED = -6
NTCAN_INVALID_PARAMETER = -7
NTCAN_INVALID_HANDLE = -8
NTCAN_TOO_MANY_HANDLES = -9
NTCAN_INIT_ERROR = -10
NTCAN_RESET_ERROR = -11
NTCAN_DRIVER_ERROR = -12
NTCAN_DLL_ALREADY_INIT = -13
NTCAN_CHANNEL_NOT_INITIALIZED = -14
NTCAN_TX_ERROR = -15
NTCAN_NO_SHAREDMEMORY = -16
NTCAN_HARDWARE_NOT_FOUND = -17
NTCAN_INVALID_NETNUMBER = -18
NTCAN_TOO_MANY_J2534_RANGES = -19
NTCAN_TOO_MANY_J2534_2_FILTERS = -20
NTCAN_DRIVER_NOT_INSTALLED = -21
NTCAN_NO_OWNER_RIGHTS = -22
NTCAN_FIRMWARE_TOO_OLD = -23
NTCAN_FIRMWARE_UNSUPPORTED = -24
NTCAN_FIRMWAREUPDATE_FAILED = -25
NTCAN_HARDWARE_NOT_SUPPORTED = -26
NTCAN_FILE_NOT_FOUND = -27
NTCAN_DEVICE_INFO_NOTAVAILABLE = -100
NTCAN_DEVICE_NOHW_ADDRESS = -101
NTCAN_NO_INTERRUPT_EVENT = -102
NTCAN_NO_INTERRUPT_EVENT_SET = -103
NTCAN_GET_MUTEX_FAILED = -104
NTCAN_NO_SHARED_MEMORY = -105
NTCAN_NET_NOT_AVAILABLE = -106
NTCAN_SETBAUDRATE_TIMEOUT = -107
NTCAN_EXE_ALREADYSTARTED = -108
NTCAN_NOTABLE_TOCREATE_SHAREDMEMORY = -109
NTCAN_HARDWARE_IN_USE = -110
NTCAN_API_NOT_RUNNING = -111
NTCAN_CHANNEL_CURR_NOT_AVAILABLE = -112
NTCAN_BUFFER_TOO_SMALL = -113
NTCAN_TOO_MANY_BRIDGE_FILTER = -114
NTCAN_HARDWARENOTACTIVE = -200
NTCAN_TOO_MANY_APPLICATIONS = -201
NTCAN_FLUSH_TIMEOUT = -202
NTCAN_NOSUCCESS = 0xFFFF0000

CANFOX_BITRATES = {
    1000000: 0,
    800000: 1,
    500000: 2,
    250000: 3,
    125000: 4,
    100000: 5,
    50000: 6,
    25000: 7,
}


# The functions defined in SIECA132.pdf
# Function name, response type, arguments tuple, error function (Optional)
_DLL_FUNCTIONS = [
    (
        "canOpen",
        c_long,
        (c_long, c_long, c_long, c_long, c_long, c_wchar_p, c_wchar_p, c_wchar_p, HANDLE),
    ),
    (
        "canOpenSH",
        c_long,
        (c_long, c_long, c_long, c_long, c_long, c_wchar_p, c_wchar_p, c_wchar_p, HANDLE),
    ),
    ("canClose", c_long, (HANDLE,)),
    ("canSetBaudrate", c_long, (HANDLE, c_long)),
    ("canSetBaudrateForce", c_long, (HANDLE, c_long)),
    ("canIsNetOwner", c_long, (HANDLE,)),
    ("canSetOwner", c_long, (c_long, HANDLE)),
    ("canGetOwner", c_long, (c_long, HANDLE)),
    ("canIdAdd", c_long, (HANDLE, c_long)),
    ("canIdAddArray", c_long, (HANDLE, POINTER(c_ubyte))),
    ("canIdDelete", c_long, (HANDLE, c_long)),
    ("canIdDeleteArray", c_long, (HANDLE,)),
    ("canIDStatus", c_long, (HANDLE, POINTER(CANIDStatusStruct))),
    ("canEnableAllIds", c_long, (HANDLE, c_bool)),
    ("canAreAllIdsEnabled", c_long, (HANDLE, POINTER(c_bool))),
    ("canSetFilterMode", c_long, (HANDLE, c_int)),
    ("canGetFilterMode", c_long, (HANDLE, c_int)),
    # "canSetFilterJ2534",
    ("canDeleteFilterJ2534", c_long, (HANDLE,)),
    # "canSetFilterJ2534_2",
    ("canDeleteFilterJ2534_2", c_long, (HANDLE,)),
    ("canSetBridgeFilter", c_long, (HANDLE, HANDLE, c_ulong, c_ulong, c_ulong)),
    # "canGetBridgeFilter",
    ("canClearBridgeFilter", c_long, (HANDLE, c_ulong, c_ulong, c_ulong, c_ulong)),
    #    ("canRead", c_long, (HANDLE, POINTER(CANMsgStruct), POINTER(c_long))),
    #    ("canRead", c_long, (HANDLE, POINTER(CANMsgBuffer(1)), POINTER(c_long))),
    #    ("canReadNoWait", c_long, (HANDLE, POINTER(CANMsgStruct), POINTER(c_long))),
    #    ("canReadNoWait", c_long, (HANDLE, POINTER(CANMsgBuffer(1)), POINTER(c_long))),
    ("canConfirmedTransmit", c_long, (HANDLE, POINTER(CANMsgStruct), POINTER(c_long))),
    ("canSend", c_long, (HANDLE, POINTER(CANMsgStruct), POINTER(c_long))),
    ("canWrite", c_long, (HANDLE, POINTER(CANMsgStruct), POINTER(c_long))),
    ("canFlush", c_long, (HANDLE, POINTER(c_long))),
    ("canStatus", c_long, (HANDLE, POINTER(CANStatusStruct))),
    # ("canGetDllInfo", c_long, (pointer, POINTER(c_void_p))),  # TODO: setup structure
    ("canGetCounter", c_long, (POINTER(CANCounterStruct),)),
    ("canGetCounterExtended", c_long, (HANDLE, POINTER(CANCounterStruct2))),
    ("canResetCounter", c_long, (HANDLE,)),
    ("canGetBusloadExtended", c_long, (HANDLE, POINTER(CANBusLoadStruct))),
    # "canGetTimeout",
    # "canSetTimeout",
    ("canBreakcanRead", c_long, (HANDLE,)),
    ("canClearBuffer", c_long, (HANDLE,)),
    ("canGetNumberOfConnectedDevices", c_long, (HANDLE,)),
    ("canGetDeviceList", c_long, (POINTER(CANInstalledDevicesStruct),)),
    ("canGetSyncTimer", c_long, (HANDLE, c_ulong, c_longlong)),
    ("canGetDeviceTimestampBase", c_long, (c_long, c_ulong)),
    ("canEnableHWExtendedId", c_long, (HANDLE, c_bool)),
    (
        "canGetCanLevel",
        c_long,
        (HANDLE, c_long, c_long, POINTER(c_double), POINTER(c_double)),
    ),
    ("canGetCanLevelHist", c_long, (HANDLE, c_bool, POINTER(CANLevelHistStruct))),
    ("canGetDiffTimeLastFrame", c_long, (HANDLE, POINTER(c_long))),
    ("canGetHWSerialNumber", c_long, (HANDLE, POINTER(c_ulong), POINTER(c_ulong))),
    ("canGetSystemTime", c_long, (POINTER(c_ulonglong), POINTER(c_ulonglong))),
    ("queryRunningVersion", c_long, (POINTER(c_ulong * 4),)),
    ("setApplicationFlags", c_long, (c_ulong,)),
    ("getApplicationFlags", c_long, (POINTER(c_ulong),)),
    ("canBlinkLED", c_long, (HANDLE, c_ulong, c_ulong, c_ulong)),
    ("canGetEepromAccess", c_long, (HANDLE, c_ulong, POINTER(c_ulong))),
    ("canReadEeprom", c_long, (HANDLE, c_long, c_long, POINTER(c_ubyte))),
    ("canWriteEeprom", c_long, (HANDLE, c_long, c_long, POINTER(c_ubyte))),
]



"""
Ctypes wrapper module for the SIE / IFM CANfox interface

Copyright (C) 2022 Matt Woodhead
"""
# standard library imports
from ctypes import c_int, c_long, c_ubyte, c_ulonglong, byref
import logging
import time
import platform
import sys

# imports from the python-can module
from can.message import Message
from can.bus import BusABC, BusState
from can.exceptions import CanOperationError, CanInitializationError, CanTimeoutError, CanInterfaceNotImplementedError
from can.ctypesutil import CLibrary, HANDLE

# from can.util import len2dlc, dlc2len

# local imports from the python-can-sontheim module
from .constants import (
    IS_PYTHON_64BIT,
    _DLL_FUNCTIONS,
    CANFOX_BITRATES,
    NTCAN_SUCCESS,
    NTCAN_RX_TIMEOUT,
    NTCAN_TX_TIMEOUT,
)
from .devices import CANfox, CANUSB, CANUSB_Legacy
from .structures import CANMsgStruct, CANMsgBuffer, CANInstalledDevicesStruct, read_struct_as_dict


try:
    # Try builtin Python 3 Windows API
    from _overlapped import CreateEvent
    from _winapi import WaitForSingleObject, WAIT_OBJECT_0, INFINITE

    HAS_EVENTS = True
except ImportError:
    try:
        # Try pywin32 package
        from win32event import CreateEvent
        from win32event import WaitForSingleObject, WAIT_OBJECT_0, INFINITE

        HAS_EVENTS = True
    except ImportError:
        # Use polling instead
        HAS_EVENTS = False


log = logging.getLogger("can.sontheim")  # Set up logging

_CANLIB = None
if sys.platform in ["win32", "cygwin"]:
    try:
        try:  # Check that python-can is being run on a 32 bit python interpreter
            assert not IS_PYTHON_64BIT
        except AssertionError as AE:
            raise CanInterfaceNotImplementedError(
                "The Sontheim API is currently only on 32 bit python interpreters"
            ) from AE
            # The SIE MT_API DLL is 32 bit only, so cannot be run from a 64 bit process
            # TODO: develop a 64 bit wrapper for the 32 bit API similar to msl-loadlib.
        _CANLIB = CLibrary("C:\\Program Files (x86)\\Sontheim\\MT_Api\\SIECA132.dll")
        for function, restype, argtypes in _DLL_FUNCTIONS:
            _CANLIB.map_symbol(function, restype, argtypes)
    except Exception as e:
        log.warning("Cannot load SIE MT_API for Sontheim: %s", e)

else:
    # Will not work on other systems, but have it importable anyway for
    # tests/sphinx
    log.warning(
        "Cannot load SIE MT_API for Sontheim does not work on %s platform, %s",
        sys.platform,
        platform.python_compiler(),
    )


def canGetSystemTime() -> int:
    """

    :raises SontheimCanOperationError:
        Raised if the Sontheim MT_API reports an error when querying the HW timestamp
    :return:
        The current system time in tenths of a millisecond (i.e divide by 10000 to get seconds)
    :rtype: int

    """

    pui64StartSysTime = c_ulonglong()
    pui64CurrSysTime = c_ulonglong()

    error_code = _CANLIB.canGetSystemTime(byref(pui64CurrSysTime), byref(pui64StartSysTime))

    if error_code != NTCAN_SUCCESS:
        raise CanOperationError("Error encountered in canGetSystemTime function call")
    return pui64CurrSysTime.value


class SontheimBus(BusABC):
    """
    A plugin for the python-can module, that allows the use of CAN interfaces that rely on the Sontheim Industrie
    Elektronik (SIE) MTAPI drivers. Currently Windows only and limited to 32 bit.
    """

    def __init__(
        self,
        channel=CANfox.CAN1,
        state=BusState.ACTIVE,
        bitrate=500000,
        *args,
        **kwargs,
    ):

        self.channel = channel
        self.channel_info = str(channel)
        self._canfox_bitrate = CANFOX_BITRATES.get(
            int(bitrate),
            CANFOX_BITRATES[500000],  # default to 500 kbit/s
        )
        self._Handle = HANDLE()
        self._bus_pc_start_time_s = None
        self._bus_hw_start_timestamp = None

        if state is BusState.ACTIVE or state is BusState.PASSIVE:
            self.state = state
        else:
            raise ValueError("BusState must be Active or Passive")

        if HAS_EVENTS:
            self._receive_event = CreateEvent(None, 0, 0, "R1")
            self._error_event = CreateEvent(None, 0, 0, "E1")

        super().__init__(channel=channel, state=state, bitrate=bitrate, *args, **kwargs)

        self._can_init(
            errors=kwargs.get(bool("errors"), True),
            echo=kwargs.get(bool("echo"), True),
            tx_timeout=kwargs.get("tx_timeout", -1),
            rx_timeout=kwargs.get("rx_timeout", -1),
        )

    def _can_init(self, errors=True, echo=False, tx_timeout=-1, rx_timeout=-1):

        # TODO: Check DLL status for DLL version - if it shows a value of zero, you need to unplug the adapter and plug it back in again to reset the driver

        error_code = _CANLIB.canOpen(
            c_long(int(self.channel)),
            c_long(errors),
            c_long(echo),
            c_long(tx_timeout),
            c_long(rx_timeout),
            "python-can",
            "R1",
            "E1",
            byref(self._Handle),
        )
        if error_code != NTCAN_SUCCESS:
            # raise CanInitializationError(
            #     "Error encountered whilst trying to open Sontheim bus interface, [Error Code: %s]" % error_code,
            # )
            raise CanInitializationError(
                f"Error encountered whilst trying to open Sontheim bus interface, [Error Code: {error_code}]",
            )

        error_code = _CANLIB.canSetBaudrate(self._Handle, c_int(self._canfox_bitrate))
        if error_code != NTCAN_SUCCESS:
            raise CanInitializationError(
                f"Error encountered whilst trying to set bus bitrate, [Error Code: {error_code}]",
            )
        error_code = _CANLIB.canSetFilterMode(self._Handle, c_int(4))
        if error_code != NTCAN_SUCCESS:
            raise CanInitializationError(
                f"Error encountered whilst trying to set bus filters, [Error Code: {error_code}]",
            )

        self._bus_pc_start_time_s = round(time.time(), 4)
        self._bus_hw_start_timestamp = canGetSystemTime() / 10000

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, new_state):
        # declare here, which is called by __init__()
        self._state = new_state  # pylint: disable=attribute-defined-outside-init

        if new_state is BusState.ACTIVE:
            pass  # TODO: change HW mode
            # self.m_objPCANBasic.SetValue(
            #     self.m_PcanHandle, PCAN_LISTEN_ONLY, PCAN_PARAMETER_OFF
            # )

        elif new_state is BusState.PASSIVE:
            pass  # TODO: change HW mode
            # When this mode is set, the CAN controller does not take part on active events
            # (eg. transmit CAN messages) but stays in a passive mode (CAN monitor),
            # in which it can analyse the traffic on the CAN bus used by a
            # PCAN channel. See also the Philips Data Sheet "SJA1000 Stand-alone CAN controller".
            # self.m_objPCANBasic.SetValue(
            #     self.m_PcanHandle, PCAN_LISTEN_ONLY, PCAN_PARAMETER_ON
            # )

    def shutdown(self):
        super().shutdown()
        _CANLIB.canClose(self._Handle)

    def _recv_internal(self, timeout):

        if HAS_EVENTS:
            # We will utilize events for the timeout handling
            timeout_ms = int(timeout * 1000) if timeout is not None else INFINITE
        elif timeout is not None:
            # Calculate max time
            end_time = time.perf_counter() + timeout

        log.debug("Trying to read a msg")

        msg_struct = CANMsgStruct()
        error_code = None
        while error_code is None:
            error_code = _CANLIB.canReadNoWait(self._Handle, byref(msg_struct), byref(c_long(1)))
            if error_code == NTCAN_RX_TIMEOUT:
                if HAS_EVENTS:
                    error_code = None
                    val = WaitForSingleObject(self._receive_event, timeout_ms)
                    if val != WAIT_OBJECT_0:
                        return None, False
                elif timeout is not None and time.perf_counter() >= end_time:
                    return None, False
                else:
                    error_code = None
                    time.sleep(0.001)
            elif error_code != NTCAN_SUCCESS:
                raise CanOperationError(
                    "Error encountered whilst trying to read bus, [Error Code: {error_code}]",
                )

        log.debug("Received a message")

        # remove bits 4 to 7 as these are reserved for other functionality
        dlc = int(msg_struct.by_len & 0x0F)

        # Use the starting timestamp
        timestamp = self._bus_pc_start_time_s + (int(msg_struct.ul_tstamp) / 10000) - self._bus_hw_start_timestamp

        frame_info = msg_struct.by_extended

        rx_msg = Message(
            timestamp=timestamp,
            arbitration_id=msg_struct.l_id,
            is_extended_id=frame_info & 2,
            is_remote_frame=msg_struct.by_remote & 1,
            is_error_frame=frame_info & 64,
            dlc=dlc,
            data=msg_struct.aby_data,
            is_fd=False,
            # bitrate_switch=bitrate_switch,
            # error_state_indicator=error_state_indicator,
        )

        return rx_msg, False

    def _recv_multiple(self, msg_buffer_length=20) -> list:

        log.debug("Trying to read multiple messages")

        msg_return_count = c_long(msg_buffer_length)
        msg_buffer = CANMsgBuffer(msg_buffer_length)
        error_code = None
        while error_code is None:
            error_code = _CANLIB.canReadNoWait(self._Handle, byref(msg_buffer), byref(msg_return_count))
            if error_code != NTCAN_SUCCESS:
                raise CanOperationError(
                    "Error encountered whilst trying to read bus, [Error Code: {error_code}]",
                )

        log.debug("Received %s message(s)", msg_return_count)

        message_list = []
        for i in range(msg_return_count.value):
            msg_struct = msg_buffer.msgs[i]
            # remove bits 4 to 7 as these are reserved for other functionality
            dlc = int(msg_struct.by_len & 0x0F)

            # Use the starting timestamp
            timestamp = self._bus_pc_start_time_s + (int(msg_struct.ul_tstamp) / 10000) - self._bus_hw_start_timestamp

            frame_info = msg_struct.by_extended

            rx_msg = Message(
                timestamp=timestamp,
                arbitration_id=msg_struct.l_id,
                is_extended_id=frame_info & 2,
                is_remote_frame=msg_struct.by_remote & 1,
                is_error_frame=frame_info & 64,
                dlc=dlc,
                data=msg_struct.aby_data,
                is_fd=False,
                # bitrate_switch=bitrate_switch,
                # error_state_indicator=error_state_indicator,
            )

            message_list.append(rx_msg)

        return message_list, False

    def send(self, msg, timeout=None):

        assert msg.dlc <= 8

        msg_struct = CANMsgStruct()

        # configure the message. ID, Data length, ID type, message type
        msg_struct.l_id = c_long(msg.arbitration_id)
        msg_struct.by_len = msg.dlc
        if msg.is_extended_id:
            msg_struct.by_extended = c_ubyte(2)  # 00000010
        else:
            msg_struct.by_extended = c_ubyte(1)  # 00000001
        if msg.is_remote_frame:
            msg_struct.by_remote = c_ubyte(1)  # 00000001
        else:
            msg_struct.by_remote = c_ubyte(0)  # 00000000

        # copy data
        for i in range(msg.dlc):
            msg_struct.aby_data[i] = msg.data[i]

        # error_code = _CANLIB.canConfirmedTransmit(self._Handle, byref(msg_struct), byref(c_long(1)))
        error_code = _CANLIB.canSend(self._Handle, byref(msg_struct), byref(c_long(1)))

        if error_code == NTCAN_TX_TIMEOUT:
            raise CanTimeoutError("Timeout whilst attempting to send message")
        if error_code != NTCAN_SUCCESS:
            raise CanOperationError(
                "Error encountered whilst trying to write to bus, [Error Code: {error_code}]",
            )

    def flush_tx_buffer(self):
        """
        This method flushes the transmit buffer to make sure all messages have been sent. Note: This is only available for use with the Sontheim CANUSB interface, and an exception will be raised if it is attempted with a different interface.

        :raises CanOperationError:
            Raised if the flush_tx_buffer method is attempted with an inteface other than the Sontheim CANUSB
        :raises CanTimeoutError:
            Raised if the operation completes because of a time out error
        :raises CanOperationError:
            Raised if a return code other than NTCAN_SUCCESS or NTCAN_TX_TIMEOUT is returned by the Sontheim API
        :return:
            None
        :rtype:
            None

        """
        try:
            assert self.channel in [
                CANUSB.CAN1,
                CANUSB.CAN2,
                CANUSB_Legacy.CAN1,
                CANUSB_Legacy.CAN2,
            ]
        except AssertionError as AE:
            raise CanOperationError(
                "The flush_tx_buffer method is only available on the sontheim CANUSB interface"
            ) from AE

        error_code = _CANLIB.canFlush(self._Handle, c_long(10000))  # ten second timeout

        if error_code == NTCAN_TX_TIMEOUT:
            raise CanTimeoutError("Timeout whilst attempting to flush TX buffer")
        if error_code != NTCAN_SUCCESS:
            raise CanOperationError(
                "Error encountered whilst trying to flush TX buffer, [Error Code: {error_code}]",
            )

    def canBlinkLED(self, blink_length_s=2):
        """

        :param blink_length_s: The length in seconds for how long the blinking is desired.
        :type blink_length_s: int

        """

        for i in range(int(blink_length_s / 0.25) + 1):
            _CANLIB.canBlinkLED(self._Handle, 1, i % 2, 5)
            time.sleep(0.25)

        error_code = _CANLIB.canBlinkLED(self._Handle, 0, i % 2, 5)
        if error_code != NTCAN_SUCCESS:
            raise CanOperationError(
                "Error encountered whilst trying to flash adpter LEDs, [Error Code: {error_code}]",
            )

    def clear_rx_buffer(self):
        """
        Clears the receive buffer in the attached CAN peripheral
        :raises CanOperationError: If an error was encountered trying to clear the buffer
        """
        error_code = _CANLIB.canClearBuffer(self._Handle)
        if error_code != NTCAN_SUCCESS:
            raise CanOperationError(
                "Error encountered whilst trying to clear the RX buffer, [Error Code: {error_code}]",
            )

    @staticmethod
    def _detect_available_configs():
        try:
            devices_struct = CANInstalledDevicesStruct()
            error_code = _CANLIB.canGetDeviceList(byref(devices_struct))
            if error_code == NTCAN_SUCCESS:
                _devices = read_struct_as_dict(devices_struct)
                if _devices:
                    return [{"interface": "sontheim", "channel": _devices["Net"]}]
            return []
        except AttributeError:
            # An AttributeError is raised when run on a 64 bit system even though the bus is not avalable
            # return an empty list to prevent issues with python-can detect_available_configs function
            return []

    def fileno(self) -> int:
        raise NotImplementedError("fileno is not implemented in the Sontheim CAN bus interfaces")

    # def __getattr__(self, item: str):
    #     return self.bus.__getattribute__(item)



# -*- coding: utf-8 -*-
"""
Created on Sat Jul  2 20:36:51 2022

sie_canfox_logger_example.py

python-can-sontheim
"""

from datetime import datetime as dt

import can
from can_sontheim import devices


def main() -> None:

    bus = can.Bus(interface="sontheim", channel=devices.CANfox.CAN1, bitrate=250000, echo=False)

    start_time = dt.today().strftime("%Y-%m-%d_%H-%M-%S")

    print(f"Connected to {bus.__class__.__name__}: {bus.channel_info}")
    print(f"Can Logger (Started on {start_time})")

    logger = can.SizedRotatingLogger(base_filename=f"{start_time}_canfox_logger.asc", max_bytes=1024000)

    try:
        while True:
            msg = bus.recv(1)
            if msg is not None:
                logger(msg)
    except KeyboardInterrupt:
        pass
    finally:
        bus.shutdown()
        logger.stop()


if __name__ == "__main__":
    main()



"""
Device definitions module for the SIE / IFM CANfox interface

Copyright (C) 2022 Matt Woodhead
"""

from collections import namedtuple


single_channel = namedtuple("single_channel", ["CAN1"])
dual_channel = namedtuple("dual_channel", ["CAN1", "CAN2"])


CANAS = dual_channel(0, 1)
CANfox = single_channel(105)
CANUSB = dual_channel(21, 22)
CANUSB_Legacy = dual_channel(24, 25)
MobiCAN = dual_channel(90, 91)
PC104Plus = dual_channel(0, 1)
PowerPCI = dual_channel(0, 1)
PowerPCIV2 = dual_channel(0, 1)
Virtual_Device_1 = dual_channel(27, 28)
Virtual_Device_2 = dual_channel(30, 31)
Virtual_Device_3 = dual_channel(33, 34)
Virtual_Device_4 = dual_channel(36, 37)
Virtual_Device_5 = dual_channel(39, 40)



"""
Test for Sontheim API Interface
"""

import ctypes
import unittest
from unittest import mock
from unittest.mock import Mock

import can
from can.bus import BusState
from can.exceptions import CanOperationError, CanInitializationError, CanTimeoutError
import can_sontheim.constants as const
from can_sontheim import SontheimBus, devices, IS_PYTHON_64BIT


skip_all_tests = IS_PYTHON_64BIT


class TestSontheimBus(unittest.TestCase):
    """unit tests for the sontheim bus"""

    @unittest.skipIf(skip_all_tests, reason="Currently requires 32 bit python interpreter only")
    def setUp(self) -> None:
        try:
            self.bus = can.Bus(interface="sontheim", channel=devices.CANfox.CAN1)
        except can.CanInterfaceNotImplementedError:
            raise unittest.SkipTest("Interface not available on this platform")

    @unittest.skipIf(skip_all_tests, reason="Currently requires 32 bit python interpreter only")
    def tearDown(self) -> None:
        if self.bus:
            self.bus.shutdown()
            self.bus = None

    @unittest.skipIf(skip_all_tests, reason="Currently requires 32 bit python interpreter only")
    def test_bus_creation(self) -> None:
        self.bus = can.Bus(bustype="sontheim")
        self.assertIsInstance(self.bus, SontheimBus)

    @unittest.skipIf(skip_all_tests, reason="Currently requires 32 bit python interpreter only")
    def test_bus_creation_state_error(self) -> None:
        with self.assertRaises(ValueError):
            can.Bus(bustype="sontheim", state=BusState.ERROR)

    @unittest.skipIf(skip_all_tests, reason="Currently requires 32 bit python interpreter only")
    def test_status(self) -> None:
        self.bus = can.Bus(bustype="sontheim")
        self.bus.status()


if __name__ == "__main__":
    unittest.main()



"""
Ctypes Structures definitions module for the SIE / IFM CANfox interface

Copyright (C) 2022 Matt Woodhead
"""

from ctypes import (
    c_char,
    c_double,
    c_int,
    c_uint,
    c_short,
    c_ushort,
    c_long,
    c_ulong,
    c_longlong,
    c_ulonglong,
    c_byte,
    c_ubyte,
    c_void_p,
    c_bool,
    byref,
)
from ctypes import Structure


MAX_NUM_APIHANDLE = 4


def read_struct_as_dict(struct):
    result = {}
    # print struct
    def get_value(value):
        if (type(value) not in [int, float, bool]) and not bool(value):
            # it's a null pointer
            value = None
        elif hasattr(value, "_length_") and hasattr(value, "_type_"):
            # Probably an array
            # print value
            value = get_array(value)
        elif hasattr(value, "_fields_"):
            # Probably another struct
            value = read_struct_as_dict(value)
        return value

    def get_array(array):
        ar = []
        for value in array:
            value = get_value(value)
            ar.append(value)
        return ar

    for f in struct._fields_:
        field = f[0]
        value = getattr(struct, field)
        # if the type is not a primitive and it evaluates to False ...
        value = get_value(value)
        result[field] = value
    return result


class CANMsgStruct(Structure):
    _fields_ = [
        ("l_id", c_long),
        ("by_len", c_ubyte),
        ("by_msg_lost", c_ubyte),
        ("by_extended", c_ubyte),
        ("by_remote", c_ubyte),
        ("aby_data", c_ubyte * 8),
        ("ul_tstamp", c_ulong),
    ]  # CMSG


def CANMsgBuffer(length: int) -> Structure:
    """
    A Helper function to return a ctypes struct class object with a can message buffer
    of the desired length.
    :param length: The number of can messages to store in the struct
    :type length: int
    :return: The ctypes array of message structs
    :rtype: Structure
    """

    class _CANMsgBuffer(Structure):
        _fields_ = [("msgs", CANMsgStruct * length)]

    if length > 1:
        return _CANMsgBuffer()
    else:
        return _CANMsgBuffer  # return the class object for symbol mapping


class CANIDStatusStruct(Structure):
    _fields_ = [
        ("aby_ID", c_ubyte * 2048),
    ]  # T_ID_ARRAY


class CANStatusStruct(Structure):
    _fields_ = [
        ("w_hw_rev", c_ushort),
        ("w_fw_rev", c_ushort),
        ("w_drv_rev", c_ushort),
        ("w_dll_rev", c_ushort),
        ("ul_board_status", c_ulong),
        ("by_board_id", c_char),
        ("w_busoffctr", c_ushort),
        ("w_errorflag", c_ushort),
        ("w_errorframectr", c_ushort),
        ("w_netctr", c_ushort),
        ("w_baud", c_ushort),
        ("ui_epld_rev", c_uint),
    ]  # CAN_IF_STATUS


# class CANDLLStruct(Structure):
#     _fields_ = [
#         ("aui_TxCounter", c_uint * 2),
#         ("aui_TxHandleCounter", c_uint * MAX_NUM_APIHANDLE),
#         ("aui_TxCounterRTR", c_uint * 2),
#         ("aui_TxHandleCounterRTR", c_uint * MAX_NUM_APIHANDLE),
#         ("aui_TxThreadCounter", c_uint * 2),
#         ("aui_TxThreadCounterRTR", c_uint * 2),
#         ("aui_RxCounter", c_uint * MAX_NUM_APIHANDLE),
#         ("aui_RxThreadCounter", c_uint * 2),
#         ("aui_RxBufferCounter", c_uint * MAX_NUM_APIHANDLE),
#         ("aui_InterfaceCtr", c_uint * MAX_NUM_APIHANDLE),
#         ("appNames", STRINGARRAY),  # TODO - work out how to implement a stringaray in python
#         ("aui_ThreadStatus", c_uint * MAX_NUM_APIHANDLE),
#         ("aui_ZugriffsCounterRead", c_uint * MAX_NUM_APIHANDLE),
#         ("aui_ZugriffsCounterWrite", c_uint * MAX_NUM_APIHANDLE),
#         ("aui_HandleNr", c_uint * MAX_NUM_APIHANDLE),
#         ("aui_NetzZuordnung", c_uint * MAX_NUM_APIHANDLE),
#         ("aui_NetzOwner", c_uint * 2),
#         ("aui_Reserve1601", c_uint * MAX_NUM_APIHANDLE),
#         ("aui_Reserve1602", c_uint * MAX_NUM_APIHANDLE),
#         ("aui_Reserve1603", c_uint * MAX_NUM_APIHANDLE),
#         ("aui_Reserve1604", c_uint * MAX_NUM_APIHANDLE),
#         ("ui_NetCount", c_uint),
#         ("ui_Reserve", c_uint),
#         ("aui_AnzahlEchoFrames", c_uint * 2),
#         ("ui_CloseZaehler", c_uint),
#         ("ui_OpenZaehler", c_uint),
#         ("ui_CloseFlag", c_uint),
#         ("ui_OpenedHandles", c_uint),
#     ]


class CANCounterStruct(Structure):
    _fields_ = [
        ("ul_timer", c_ulong),
        ("ul_rxframectr1", c_ulong),
        ("ul_txframectr1", c_ulong),
        ("ul_txremframectr1", c_ulong),
        ("ul_txdatabytectr1", c_ulong),
        ("ul_rxdatabytectr1", c_ulong),
        ("reserved0", c_ulong),
        ("ul_rxframectr2", c_ulong),
        ("l_txframectr2", c_ulong),
        ("ul_txremframectr2", c_ulong),
        ("ul_txdatabytectr2", c_ulong),
        ("ul_rxdatabytectr2", c_ulong),
        ("ul_CAN1ctrlflags", c_ulong),
        ("ul_CAN2ctrlflags", c_ulong),
        ("ul_errframectr1", c_ulong),
        ("ul_errframectr2", c_ulong),
    ]  # CTRDATA


class CANCounterStruct2(Structure):
    _fields_ = [
        ("ul_timer", c_ulong),
        ("ul_rxframectr1", c_ulong),
        ("ul_txframectr1", c_ulong),
        ("ul_txremframectr1", c_ulong),
        ("ul_txdatabytectr1", c_ulong),
        ("ul_rxdatabytectr1", c_ulong),
        ("reserved0", c_ulong),
        ("ul_rxframectr2", c_ulong),
        ("l_txframectr2", c_ulong),
        ("ul_txremframectr2", c_ulong),
        ("ul_txdatabytectr2", c_ulong),
        ("ul_rxdatabytectr2", c_ulong),
        ("ul_CAN1ctrlflags", c_ulong),
        ("ul_CAN2ctrlflags", c_ulong),
        ("ul_errctr1", c_ulong),
        ("ul_errctr2", c_ulong),
        ("ul_rxremframectr1", c_ulong),
        ("ul_erxframectr1", c_ulong),
        ("ul_etxframectr1", c_ulong),
        ("ul_etxremframectr1", c_ulong),
        ("ul_erxremframectr1", c_ulong),
        ("ul_rxremframectr2", c_ulong),
        ("ul_erxframectr2", c_ulong),
        ("ul_etxframectr2", c_ulong),
        ("ul_etxremframectr2", c_ulong),
        ("ul_erxremframectr2", c_ulong),
    ]  # CTRDATA2


# class CANTimeoutTypeStruct(Structure):  # TODO: Should be ENUM
#     _fields_ = [
#         ("aby_ID", c_ubyte * 2048),
#     ]  # T_ID_ARRAY


class CANInstalledDevicesStruct(Structure):
    _fields_ = [
        ("Net", c_int),
        ("Name", c_char * 20),
        ("ul_Status", c_ulong),
        ("ul_Features", c_ulong),
        ("Reserved", c_long * 18),
    ]  # T_DeviceList


# class CANTimeoutTypeStruct(Structure):  # TODO: Should be ENUM
#     _fields_ = [
#         ("canLevelDominant", c_uint),
#         ("canLevelRecessive", c_uint),
#     ]  # T_ID_ARRAY


class CANLevelHistStruct(Structure):
    _fields_ = [
        ("ulMeasureCountAll", c_ulong),
        ("arr_LevelLow", c_ulong * 50),
        ("arr_LevelHigh", c_ulong * 50),
        ("ulMeasureCountLowRez", c_ulong),
        ("dAvgLevelLowRez", c_double),
        ("dVariLevelLowRez", c_double),
        ("ulMeasureCountHighRez", c_ulong),
        ("dAvgLevelHighRez", c_double),
        ("dVariLevelHighRez", c_double),
        ("ulMeasureCountLowDom", c_ulong),
        ("dAvgLevelLowDom", c_double),
        ("dVariLevelLowDom", c_double),
        ("ulMeasureCountHighDom", c_ulong),
        ("dAvgLevelHighDom", c_double),
        ("dVariLevelHighDom", c_double),  # Variance of average level of dominant CAN HIGH
    ]  # CANLEVEL_HIST


# typedef enum
# {
# filterMode_standard = 0,
# filterMode_j2534 = 1,
# filterMode_extended = 2,
# filterMode_j2534_2 = 3,
# filterMode_nofilter = 4
# } T_FILTER_MODE;


# typedef enum
# {
# j2534Mode_excl = 0,
# j2534Mode_incl = 1
# } T_J2534_MODE;


class CANBusLoadStruct(Structure):
    _fields_ = [
        ("ul_intervall", c_ulong),
        ("ul_stdframectr", c_ulong),
        ("ul_stdframebytectr", c_ulong),
        ("ul_extframectr", c_ulong),
        ("ul_extframebytectr", c_ulong),
        ("ObjectTime", c_double),
        ("ul_load", c_ulong),
    ]  # BUSLOAD


class CANBridgeFilterStruct(Structure):
    _fields_ = [
        ("ulDestNet", c_ulong),
        ("ulPattern", c_ulong),
        ("ulMask", c_ulong),
        ("ulReserved", c_ulong),
    ]  # BRIDGE_FILTER



# -*- coding: utf-8 -*-
"""
Created on Sat Jul  2 20:36:51 2022

sie_canfox_receive_multiple_example.py

python-can-sontheim
"""

from time import sleep
from pprint import pprint

import can
from can_sontheim import devices, _canlib


def main() -> None:

    print(f"System has events: {_canlib.HAS_EVENTS}")

    with can.Bus(interface="sontheim", channel=devices.CANfox.CAN1, bitrate=250000, echo=False) as bus:
        try:
            no_recv = True
            bus.clear_rx_buffer()
            while no_recv:
                msg = bus.recv(1)
                if msg is not None:
                    break
            print(msg)
            sleep(0.1)
            msgs = bus._recv_multiple(500)
            sleep(0.1)
            msgs = bus._recv_multiple(500)
            if (msgs is not None) or (msgs != []):
                pprint(msgs[0])
                print(len(msgs[0]))
        except KeyboardInterrupt:
            bus.shutdown()
        finally:
            bus.shutdown()


if __name__ == "__main__":

    main()



# -*- coding: utf-8 -*-
"""
Created on Sat Jul  2 20:36:51 2022

sie_canfox_tx_example.py

python-can-sontheim
"""

import can
from can import Message
from can_sontheim import devices
from threading import Timer
from time import ctime


DONE = False

with can.Bus(interface="sontheim", channel=devices.CANfox.CAN1, bitrate=250000, echo=False) as bus:

    def timeout():
        global DONE
        DONE = True

    msg = Message(
        arbitration_id=0xC0FFEF,
        is_extended_id=True,
        data=[0xDE, 0xAD, 0xBE, 0xEF, 0xDE, 0xAD, 0xBE, 0xEF],
    )

    task = bus.send_periodic(msg, 1)
    t = Timer(10, timeout)
    t.start()
    print(f"TX started: {ctime()}")

    while not DONE:
        pass
    else:
        task.stop()
        print(f"TX Done: {ctime()}")



# -*- coding: utf-8 -*-
"""
Created on Sun Dec  4 17:12:37 2022

init.py

python-can-sontheim
"""

from .version import __version__
from ._canlib import SontheimBus
from .constants import IS_PYTHON_64BIT
